/****************************************************************************
 * Functions to load fonts.
 * Copyright (C) 2004 Joe Wingbermuehle
 ****************************************************************************/

#include "jwm.h"

static const char *DEFAULT_BORDER_FONT = "-*-times-*-r-*-*-*-*-*-*-*-*-*-*";
static const char *DEFAULT_TRAY_FONT = "-*-times-*-r-*-*-*-*-*-*-*-*-*-*";
static const char *DEFAULT_MENU_FONT = "-*-times-*-r-*-*-*-*-*-*-*-*-*-*";

static char *borderFontName = NULL;
static char *trayFontName = NULL;
static char *menuFontName = NULL;

XFontStruct *borderFont = NULL;
XFontStruct *trayFont = NULL;
XFontStruct *menuFont = NULL;

/* Data for antialiasing */
static unsigned long *inputData = NULL;
static unsigned long *outputData = NULL;
static XImage *inputImage = NULL;
static XImage *outputImage = NULL;
static Pixmap buffer = None;
static GC bufferGC;
static int antialias = 0;

static unsigned long white;
static unsigned long black;

static void Antialias(int width, int height, RampType ramp);

/****************************************************************************
 ****************************************************************************/
void InitializeFonts() {

	white = WhitePixel(display, DefaultScreen(display));
	black = BlackPixel(display, DefaultScreen(display));

	borderFont = JXLoadQueryFont(display, borderFontName);
	if(!borderFont) {
		borderFont = JXLoadQueryFont(display, DEFAULT_BORDER_FONT);
		if(!borderFont) {
			FatalError("Could not load the default border font");
		}
	}

	trayFont = JXLoadQueryFont(display, trayFontName);
	if(!trayFont) {
		trayFont = JXLoadQueryFont(display, DEFAULT_TRAY_FONT);
		if(!trayFont) {
			FatalError("Could not load the default tray font");
		}
	}

	menuFont = JXLoadQueryFont(display, menuFontName);
	if(!menuFont) {
		menuFont = JXLoadQueryFont(display, DEFAULT_MENU_FONT);
		if(!menuFont) {
			FatalError("Could not load the default menu font");
		}
	}


}

/****************************************************************************
 ****************************************************************************/
void SetBorderFont(const char *fontName) {
	if(borderFontName) {
		Release(borderFontName);
		borderFontName = NULL;
	}
	if(fontName) {
		borderFontName = Allocate(strlen(fontName) + 1);
		strcpy(borderFontName, fontName);
	}
}

/****************************************************************************
 ****************************************************************************/
void SetTrayFont(const char *fontName) {
	if(trayFontName) {
		Release(trayFontName);
		trayFontName = NULL;
	}
	if(fontName) {
		trayFontName = Allocate(strlen(fontName) + 1);
		strcpy(trayFontName, fontName);
	}
}

/****************************************************************************
 ****************************************************************************/
void SetMenuFont(const char *fontName) {
	if(menuFontName) {
		Release(menuFontName);
		menuFontName = NULL;
	}
	if(fontName) {
		menuFontName = Allocate(strlen(fontName) + 1);
		strcpy(menuFontName, fontName);
	}
}

/****************************************************************************
 ****************************************************************************/
void SetAntialias(int v) {
	antialias = v;
}

/****************************************************************************
 ****************************************************************************/
void Antialias(int width, int height, RampType ramp) {

	int x, y;
	int offset;
	float value;
	int line = inputImage->width;

	for(y = 1; y < height - 1; y++) {
		for(x = 1; x < width - 1; x++) {

			offset = y * line + x;

			if(   inputData[offset + 1] != inputData[offset - 1]
				&& inputData[offset + line + 1] != inputData[offset - line - 1]
				&& inputData[offset + line - 1] != inputData[offset - line + 1]
				&& inputData[offset + line] != inputData[offset - line]) {

				if(inputData[offset] == black) {
					value = 0.25;
				} else {
					value = 0.0;
				}
				if(inputData[offset + 1] == black) {
					value += 0.1875;
				}
				if(inputData[offset - 1] == black) {
					value += 0.1875;
				}
				if(inputData[offset + line] == black) {
					value += 0.1875;
				}
				if(inputData[offset - line] == black) {
					value += 0.1875;
				}

				outputData[offset] = ramps[ramp][(int)(value * 255.0)];

			} else {
				if(inputData[offset] == black) {
					outputData[offset] = ramps[ramp][255];
				} else {
					outputData[offset] = ramps[ramp][0];
				}
			}

		}
	}

}

/****************************************************************************
 ****************************************************************************/
void RenderString(Drawable d, GC g, XFontStruct *font,
	RampType ramp, int x, int y, int width, const char *str) {

	int len, h, w;

	Assert(font);
	Assert(str);

	len = strlen(str);

	if(!antialias) {
		if(ramp == RAMP_ACTIVE) {
			JXSetForeground(display, g, colors[COLOR_ACTIVE_BORDER_TEXT]);
		} else {
			JXSetForeground(display, g, colors[COLOR_BORDER_TEXT]);
		}
		JXSetFont(display, g, font->fid);
		JXDrawString(display, d, g, x, y + font->ascent, str, len);
		return;
	}

	h = Min(font->ascent + font->descent, rootHeight) + 2;
	w = Min(JXTextWidth(font, str, len), width) + 2;

	if(!inputData || h > inputImage->height || w > inputImage->width) {

		if(inputData) {

			Release(inputData);
			inputImage->data = NULL;
			JXDestroyImage(inputImage);

			Release(outputData);
			outputImage->data = NULL;
			JXDestroyImage(outputImage);

			JXFreeGC(display, bufferGC);
			JXFreePixmap(display, buffer);

		}

		buffer = JXCreatePixmap(display, rootWindow, w, h,
			DefaultDepth(display, DefaultScreen(display)));
		bufferGC = JXCreateGC(display, buffer, 0, NULL);

		inputData = Allocate(w * h * sizeof(unsigned long));
		inputImage = JXCreateImage(display,
			DefaultVisual(display, DefaultScreen(display)),
			DefaultDepth(display, DefaultScreen(display)),
			ZPixmap, 0, (char*)inputData, w, h, 32, 0);

		outputData = Allocate(w * h * sizeof(unsigned long));
		outputImage = JXCreateImage(display,
			DefaultVisual(display, DefaultScreen(display)),
			DefaultDepth(display, DefaultScreen(display)),
			ZPixmap, 0, (char*)outputData, w, h, 32, 0);

	}

	JXSetForeground(display, bufferGC, white);
	JXFillRectangle(display, buffer, bufferGC, 0, 0, w, h);

	JXSetForeground(display, bufferGC, black);
	JXSetFont(display, bufferGC, font->fid);
	JXDrawString(display, buffer, bufferGC, 1,
		1 + font->ascent, str, len);

	JXGetSubImage(display, buffer, 0, 0, w, h, AllPlanes, ZPixmap,
		inputImage, 0, 0);

	Antialias(w, h, ramp);

	JXPutImage(display, d, g, outputImage, 1, 1, x, y, w - 2, h - 2);

}

/****************************************************************************
 ****************************************************************************/
void ReleaseFonts() {
	if(borderFont) {
		JXFreeFont(display, borderFont);
		borderFont = NULL;
	}
	if(trayFont) {
		JXFreeFont(display, trayFont);
		trayFont = NULL;
	}
	if(menuFont) {
		JXFreeFont(display, menuFont);
		menuFont = NULL;
	}

	if(borderFontName) {
		Release(borderFontName);
		borderFontName = NULL;
	}
	if(trayFontName) {
		Release(trayFontName);
		trayFontName = NULL;
	}
	if(menuFontName) {
		Release(menuFontName);
		menuFontName = NULL;
	}

	if(inputData) {
		Release(inputData);
		inputImage->data = NULL;
		JXDestroyImage(inputImage);

		Release(outputData);
		outputImage->data = NULL;
		JXDestroyImage(outputImage);

		JXFreeGC(display, bufferGC);
		JXFreePixmap(display, buffer);
		buffer = None;
		inputData = NULL;
	}

}


