/****************************************************************************
 * Functions to handle hints.
 * Copyright (C) 2004 Joe Wingbermuehle
 ****************************************************************************/

#include "jwm.h"

typedef struct {
	Atom *atom;
	const char *name;
} ProtocolNode;

/* The Basics */
Atom protWMState;
Atom protWMProtocols;
Atom protWMDeleteWindow;
Atom protWMTakeFocus;
Atom protWMLocaleName;
Atom protWMCommand;
Atom protWMChangeState;

/* WM Spec */
Atom protNetSupported;
Atom protNetNumberOfDesktops;
Atom protNetDesktopGeometry;
Atom protNetDesktopViewport;
Atom protNetCurrentDesktop;
Atom protNetActiveWindow;
Atom protNetWorkarea;
Atom protNetSupportingWMCheck;
Atom protNetWMDesktop;
Atom protNetWMState;
Atom protNetWMStateSticky;
Atom protNetWMName;
Atom protNetWMIcon;
Atom protNetWMStateSticky;

/* GNOME Hints */
Atom protWinLayer;
Atom protWinState;
Atom protWinWorkspaceCount;
Atom protWinWorkspace;
Atom protWinSupportingWMCheck;
Atom protWinProtocols;

/* MWM Hints */
#ifdef USE_MWM_HINTS
Atom protMWMHints;
#endif

static const ProtocolNode gnomeProtocols[] = {
	{ &protWinLayer,              "_WIN_LAYER"                },
	{ &protWinState,              "_WIN_STATE"                },
	{ &protWinWorkspace,          "_WIN_WORKSPACE"            },
	{ &protWinWorkspaceCount,     "_WIN_WORKSPACE_COUNT"      },
	{ &protWinSupportingWMCheck,  "_WIN_SUPPORTING_WM_CHECK"  },
	{ &protWinProtocols,          "_WIN_PROTOCOLS"            },
	{ NULL,                       NULL                        }
};

static const ProtocolNode netProtocols[] = {
	{ &protNetSupported,          "_NET_SUPPORTED"            },
	{ &protNetNumberOfDesktops,   "_NET_NUMBER_OF_DESKTOPS"   },
	{ &protNetDesktopGeometry,    "_NET_DESKTOP_GEOMETRY"     },
	{ &protNetDesktopViewport,    "_NET_DESKTOP_VIEWPORT"     },
	{ &protNetCurrentDesktop,     "_NET_CURRENT_DESKTOP"      },
	{ &protNetActiveWindow,       "_NET_ACTIVE_WINDOW"        },
	{ &protNetWorkarea,           "_NET_WORKAREA"             },
	{ &protNetSupportingWMCheck,  "_NET_SUPPORTING_WM_CHECK"  },
	{ &protNetWMDesktop,          "_NET_WM_DESKTOP"           },
	{ &protNetWMState,            "_NET_WM_STATE"             },
	{ &protNetWMName,             "_NET_WM_NAME"              },
	{ &protNetWMIcon,             "_NET_WM_ICON"              },
	{ &protNetWMStateSticky,      "_NET_WM_STATE_STICKY"      },
	{ NULL,                       NULL                        }
};

#ifdef USE_MWM_HINTS
static const ProtocolNode mwmProtocols[] = {
	{ &protMWMHints,              _XA_MWM_HINTS               },
	{ NULL,                       NULL                        }
};
#endif

Atom protCompoundText;

static const ProtocolNode protocols[] = {
	{ &protWMState,               "WM_STATE"                  },
	{ &protWMProtocols,           "WM_PROTOCOLS"              },
	{ &protWMDeleteWindow,        "WM_DELETE_WINDOW"          },
	{ &protWMTakeFocus,           "WM_TAKE_FOCUS"             },
	{ &protWMLocaleName,          "WM_LOCALE_NAME"            },
	{ &protWMCommand,             "WM_COMMAND"                },
	{ &protWMChangeState,         "WM_CHANGE_STATE"           },
{ &protCompoundText, "COMPOUND_TEXT" },
	{ NULL,                       NULL                        }
};

static void ReadWMTransientFor(ClientNode *np);
static void ReadWinState(ClientNode *np);
static void ReadNetWMState(ClientNode *np);

/****************************************************************************
 ****************************************************************************/
void InitializeProtocols() {
	long array[128];
	Atom *supported;
	int x;

	for(x = 0; protocols[x].atom; x++) {
		*protocols[x].atom = JXInternAtom(display, protocols[x].name, False);
	}

	for(x = 0; netProtocols[x].atom; x++) {
		*netProtocols[x].atom = JXInternAtom(display,
			netProtocols[x].name, False);
	}
	supported = Allocate(sizeof(Atom) * x);
	for(x = 0; netProtocols[x].atom; x++) {
		supported[x] = *netProtocols[x].atom;
	}
	JXChangeProperty(display, rootWindow, protNetSupported,
		XA_ATOM, 32, PropModeReplace,
		(unsigned char*)supported, x);
	Release(supported);

	for(x = 0; gnomeProtocols[x].atom; x++) {
		*gnomeProtocols[x].atom = JXInternAtom(display,
			gnomeProtocols[x].name, False);
	}
	supported = Allocate(sizeof(Atom) * x);
	for(x = 0; gnomeProtocols[x].atom; x++) {
		supported[x] = *gnomeProtocols[x].atom;
	}
	JXChangeProperty(display, rootWindow, protWinProtocols,
		XA_ATOM, 32, PropModeReplace,
		(unsigned char*)supported, x);
	Release(supported);

#ifdef USE_MWM_HINTS
	for(x = 0; mwmProtocols[x].atom; x++) {
		*mwmProtocols[x].atom = JXInternAtom(display,
			mwmProtocols[x].name, False);
	}
#endif

	SetCardinalAtom(rootWindow, protNetNumberOfDesktops, desktopCount);

	array[0] = rootWidth;
	array[1] = rootHeight;
	JXChangeProperty(display, rootWindow, protNetDesktopGeometry,
		XA_CARDINAL, 32, PropModeReplace,
		(unsigned char*)array, 2);

	array[0] = 0;
	array[1] = 0;
	JXChangeProperty(display, rootWindow, protNetDesktopViewport,
		XA_CARDINAL, 32, PropModeReplace,
		(unsigned char*)array, 2);

	for(x = 0; x < desktopCount; x++) {
		array[x * 4 + 0] = 0;
		array[x * 4 + 1] = 0;
		array[x * 4 + 2] = rootWidth;
		array[x * 4 + 3] = rootHeight;
	}
	JXChangeProperty(display, rootWindow, protNetWorkarea,
		XA_CARDINAL, 32, PropModeReplace,
		(unsigned char*)array, desktopCount * 4);

	SetWindowAtom(rootWindow, protNetSupportingWMCheck, trayWindow);
	SetWindowAtom(trayWindow, protNetSupportingWMCheck, trayWindow);

	SetCardinalAtom(rootWindow, protWinSupportingWMCheck, trayWindow);
	SetCardinalAtom(trayWindow, protWinSupportingWMCheck, trayWindow);

	SetCardinalAtom(rootWindow, protWinWorkspaceCount, desktopCount);

}

/****************************************************************************
 ****************************************************************************/
void ReadCurrentDesktop() {
	CARD32 temp;

	currentDesktop = 0;

	if(GetCardinalAtom(rootWindow, protNetCurrentDesktop, &temp)) {
		ChangeDesktop(temp);
	} else if(GetCardinalAtom(rootWindow, protWinWorkspace, &temp)) {
		ChangeDesktop(temp);
	} else {
		ChangeDesktop(0);
	}

}

/****************************************************************************
 ****************************************************************************/
void ReadClientProtocols(ClientNode *np) {

	/* Read the standard hints */
	ReadWMName(np);
	ReadWMProtocols(np);
	ReadWMNormalHints(np);
	ReadWMTransientFor(np);

	/* Read WM hints */
	ReadNetWMDesktop(np);
	ReadNetWMState(np);

	/* Read GNOME hints */
	ReadWinState(np);
	ReadWinLayer(np);
}

/****************************************************************************
 ****************************************************************************/
void ReadNetWMState(ClientNode *np) {
	unsigned long count, x;
	int status;
	unsigned long extra;
	Atom realType;
	int realFormat;
	unsigned char *temp;
	Atom *state;

	status = JXGetWindowProperty(display, np->window, protNetWMDesktop,
		0, 32, False, XA_ATOM, &realType, &realFormat, &count,
		&extra, &temp);

	if(status == Success && count > 0) {
		state = (unsigned long*)temp;
		for(x = 0; x < count; x++) {
			if(state[x] == protNetWMStateSticky) {
				SetClientSticky(np, 1, 0);
			}
		}
	}

	if(temp) {
		JXFree(temp);
	}
}

/****************************************************************************
 ****************************************************************************/
void ReadNetWMDesktop(ClientNode *np) {
	CARD32 temp;

	Assert(np);

	if(GetCardinalAtom(np->window, protNetWMDesktop, &temp)) {
		if(temp == 0xFFFFFFFF) {
			SetClientSticky(np, 1, 0);
		} else {
			np->statusFlags &= ~STAT_STICKY;
			SetClientDesktop(np, temp, 0);
		}
	}

}

/****************************************************************************
 ****************************************************************************/
void ReadWinState(ClientNode *np) {
	CARD32 flags;

	Assert(np);

	if(GetCardinalAtom(np->window, protWinState, &flags)) {
		if(flags & WIN_STATE_STICKY) {
			SetClientSticky(np, 1, 0);
		}
		if(flags & WIN_STATE_HIDDEN) {
			np->statusFlags |= STAT_NOLIST;
		}
	}

}

/****************************************************************************
 ****************************************************************************/
void ReadWinLayer(ClientNode *np) {
	CARD32 temp;

	Assert(np);

	if(GetCardinalAtom(np->window, protWinLayer, &temp)) {
		np->layer = temp;
	}

}

/****************************************************************************
 ****************************************************************************/
void ReadWMName(ClientNode *np) {
	unsigned long count;
	int status;
	unsigned long extra;
	Atom realType;
	int realFormat;

	Assert(np);

	if(np->name) {
		JXFree(np->name);
	}
	JXFetchName(display, np->window, &np->name);

	if(!np->name) {
		status = JXGetWindowProperty(display, np->window, XA_WM_NAME,
			0, 1024, False, protCompoundText, &realType, &realFormat, &count,
			&extra, (unsigned char**)&np->name);
	}

}

/****************************************************************************
 ****************************************************************************/
void ReadWMProtocols(ClientNode *np) {
	unsigned long count, x;
	int status;
	unsigned long extra;
	Atom realType;
	int realFormat;
	unsigned char *temp;
	Atom *p;

	Assert(np);

	np->protocols = PROT_NONE;

	status = JXGetWindowProperty(display, np->window, protWMProtocols,
		0, 32, False, XA_ATOM, &realType, &realFormat, &count, &extra, &temp);
	p = (Atom*)temp;

	if(status != Success || !p) {
		return;
	}
	if(count <= 0) {
		JXFree(p);
		return;
	}

	for(x = 0; x < count; x++) {
		if(p[x] == protWMDeleteWindow) {
			np->protocols |= PROT_DELETE;
		} else if(p[x] == protWMTakeFocus) {
			np->protocols |= PROT_TAKE_FOCUS;
		}
	}

	JXFree(p);
}

/****************************************************************************
 ****************************************************************************/
void ReadWMTransientFor(ClientNode *np) {
	Window w;
	int status;

	Assert(np);

	status = JXGetTransientForHint(display, np->window, &w);
	if(status) {
		np->owner = w;
	}

}

/****************************************************************************
 ****************************************************************************/
void ReadWMNormalHints(ClientNode *np) {
	XSizeHints hints;
	long temp;

	Assert(np);

	if(!JXGetWMNormalHints(display, np->window, &hints, &temp)) {
		np->sizeFlags = 0;
	} else {
		np->sizeFlags = hints.flags;
	}

	if(np->sizeFlags & PResizeInc) {
		np->xinc = Max(1, hints.width_inc);
		np->yinc = Max(1, hints.height_inc);
	} else {
		np->xinc = 1;
		np->yinc = 1;
	}

	if(np->sizeFlags & PMinSize) {
		np->minWidth = Max(0, hints.min_width);
		np->minHeight = Max(0, hints.min_height);
	} else {
		np->minWidth = 1;
		np->minHeight = 1;
	}

	if(np->sizeFlags & PMaxSize) {
		np->maxWidth =  hints.max_width;
		np->maxHeight = hints.max_height;
		if(np->maxWidth <= 0) {
			np->maxWidth = rootWidth;
		}
		if(np->maxHeight <= 0) {
			np->maxHeight = rootHeight;
		}
	} else {
		np->maxWidth = rootWidth;
		np->maxHeight = rootHeight;
	}

	if(np->sizeFlags & PBaseSize) {
		np->baseWidth = hints.base_width;
		np->baseHeight = hints.base_height;
	} else if(np->sizeFlags & PMinSize) {
		np->baseWidth = np->minWidth;
		np->baseHeight = np->minHeight;
	} else {
		np->baseWidth = 0;
		np->baseHeight = 0;
	}

	if(np->sizeFlags & PAspect) {
		np->aspect.minx = hints.min_aspect.x;
		np->aspect.miny = hints.min_aspect.y;
		np->aspect.maxx = hints.max_aspect.x;
		np->aspect.maxy = hints.max_aspect.y;
		if(np->aspect.minx < 1) {
			np->aspect.minx = 1;
		}
		if(np->aspect.miny < 1) {
			np->aspect.miny = 1;
		}
		if(np->aspect.maxx < 1) {
			np->aspect.maxx = 1;
		}
		if(np->aspect.maxy < 1) {
			np->aspect.maxy = 1;
		}
	}

	if(np->sizeFlags & PWinGravity) {
		np->gravity = hints.win_gravity;
	} else {
		np->gravity = 1;
	}
}

/****************************************************************************
 ****************************************************************************/
int GetCardinalAtom(Window window, Atom atom, CARD32 *value) {
	unsigned long count;
	int status;
	unsigned long extra;
	Atom realType;
	int realFormat;
	unsigned char *data;

	Assert(value);

	status = JXGetWindowProperty(display, window, atom, 0, 1, False,
		XA_CARDINAL, &realType, &realFormat, &count, &extra, &data);

	if(data) {
		if(status == Success && count == 1) {
			*value = *(CARD32*)data;
		}
		JXFree(data);
	}

	return (status == Success && count == 1);

}

/****************************************************************************
 ****************************************************************************/
int GetWindowAtom(Window window, Atom atom, CARD32 *value) {
	unsigned long count;
	int status;
	unsigned long extra;
	Atom realType;
	int realFormat;
	unsigned char *data;

	Assert(value);

	status = JXGetWindowProperty(display, window, atom, 0, 1, False,
		XA_WINDOW, &realType, &realFormat, &count, &extra, &data);

	if(data) {
		if(status == Success && count == 1) {
			*value = *(CARD32*)data;
		}
		JXFree(data);
	}

	return (status == Success && count == 1);

}

/****************************************************************************
 ****************************************************************************/
void SetCardinalAtom(Window window, Atom atom, CARD32 value) {

	JXChangeProperty(display, window, atom, XA_CARDINAL, 32,
		PropModeReplace, (unsigned char*)&value, 1);

}

/****************************************************************************
 ****************************************************************************/
void SetWindowAtom(Window window, Atom atom, CARD32 value) {

	JXChangeProperty(display, window, atom, XA_WINDOW, 32,
		PropModeReplace, (unsigned char*)&value, 1);

}


